from collections import defaultdict
from pdsketch import PDPoint

class Diagram():
    """
    A class to store persistence diagrams.
    Main input is a set of points in the persistence plane with optional multiplicity.
    This implementation is a multiset.
    """

    def __init__(self, points=(), mass:list() = None):
        """
        Parameters
        ----------
        points: list
            The points in the the persistence diagram
        mass: list
            The corresponding multiplicity of points
        """
        if mass == None:
            mass = [1]*len(points)
        else:
            if len(mass) != len(points):
                raise ValueError("The lengths of mass and points should be " +
                "the same.")
        self.mass = defaultdict(int)
        self._diagonal = PDPoint([0,0])
        self.total_mass = 0
        for i, p in enumerate(points):
            self.add(PDPoint(p), mass[i])

    def points(self):
        """
        Return a list of points in the diagram.
        """
        return list(self.mass)

    def add(self, point, mass: int = 1):
        """
        Add a point to the diagram with multiplicity.
        """
        if not isinstance(point, PDPoint):
            point = PDPoint(point)
        if point.isdiagonalpoint():
            point = self._diagonal
        self.mass[point] += mass
        self.total_mass += mass
        if self.mass[point] <= 0:
            self.remove(point)

    def remove(self, point: PDPoint):
        """
        Remove a point from the diagram.
        """
        if point in self:
            del self.mass[point]
        else:
            raise KeyError("Point is not in the diagram")

    def clear(self):
        """
        Remove all points from the diagram.
        """
        self.mass.clear()

    def get_point_mass_lists(self):
        """
        Returns a tuple of lists.
        The first list contains all points in the diagram in PDPoint format.
        The second list contains all corresponding multiplicities.

        Note: The diagonal is ignored.
        """
        points = []
        masses = []
        for p in self.mass:
            if p != self._diagonal:
                points.append(p)
                masses.append(self.mass[p])
        return points, masses

    def load_from_file(filename:str):
        """
        Loads a diagram from a text file.
        Format for ith line:
        `b_i d_i; mass_i`
        where `(b_i, d_i)` is the ith point in the diagram with
        multiplicity `mass_i`.
        The mass is optional.
        """
        D = Diagram()
        with open(filename, 'r') as d:
            for line in d:
                data = line.rstrip().split("; ")
                point = PDPoint.fromstring(data[0])
                mass = int(data[1]) if len(data) >= 2 else 1
                D.add(point, mass)
        return D

    def __str__(self):
        """
        Convert the diagram to a string.
        This is the same string format that is used for saving to a file.
        """
        return '\n'.join(f'{str(p)}; {str(self.mass[p])}' for p in self)

    def save_to_file(self, filename:str = "diagram.txt"):
        """
        Save current diagram to a text file.
        Format for ith line:
        `b_i d_i; mass_i`
        where `(b_i, d_i)` is the ith point in the diagram with
        multiplicity `mass_i`.
        """
        with open(filename, 'w') as f:
            f.write(str(self))

    def __iter__(self):
        return iter(self.mass)

    def __len__(self):
        """
        Return the number of distinct points off the diagonal.
        Note: The diagonal is ignored when computing the length.
        """
        if self._diagonal in self.mass:
            return len(self.mass)-1
        else:
            return len(self.mass)

    def __contains__(self, point: PDPoint):
        """
        Test if a point is in the diagram.
        """
        return point in self.mass

    def __eq__(self, other):
        """
        Tests equality of diagrams including multiplicity.
        That is, two diagrams are equal if they have the same points and
        those points have the same multiplicity.
        """
        return self.mass == other.mass



from pdsketch import Diagram, PDPoint
from pdsketch.pdpointviz import PDPointViz
from ds2viz.element import Line
from ds2viz.canvas import svg_plus_pdf
from pdsketch.sketch_style import sketch_style

class DiagramViz():
    """
    Class to depict a persistence diagram using ds2viz.
    Stores PDPoints of a diagram as PDPointViz objects and draws them when asked to.
    """
    def __init__(self, diagram:Diagram, style:str = None, plot_size:int = 0):
        """
        Parameters
        ----------
        diagram : Diagram
            The PD to be plotted
        style : str
            Important only when plotting bottleneck matchings
        plot_size : int
            Important only when plotting bottleneck matchings
        """
        self.plot_size = max(p[1] for p in diagram)+5 if not plot_size else plot_size
        self.points = [PDPointViz(p, diagram.mass[p], self.plot_size, style) for p in diagram]

    def draw_skeleton(self, canvas):
        # Draw diagonal and boundaries of persistence diagram
        Line((0,self.plot_size), (self.plot_size,0), style='_diagonal', stylesheet=sketch_style).draw(canvas)
        Line((0,0), (self.plot_size,0), style='_boundary', stylesheet=sketch_style).draw(canvas)
        Line((self.plot_size,0), (self.plot_size,self.plot_size), style='_boundary', stylesheet=sketch_style).draw(canvas)
        Line((self.plot_size,self.plot_size), (0,self.plot_size), style='_boundary', stylesheet=sketch_style).draw(canvas)
        Line((0,self.plot_size), (0,0), style='_boundary', stylesheet=sketch_style).draw(canvas)

    def draw_points(self, canvas):
        # Draw points of the persistence diagrame
        [p.draw(canvas) for p in self.points]

    def draw(self, canvas):
        self.draw_skeleton(canvas)
        self.draw_points(canvas)

    # def draw(self, filename:str):
    #     # Draw both, points and boundaries
    #     with svg_plus_pdf(self.plot_size, self.plot_size, filename, sketch_style) as canvas:
    #         self.draw_skeleton(canvas)
    #         self.draw_points(canvas)



class PDPoint:
    """
    A class to describe points in the persistence plane.
    """
        
    def __init__(self, coords):
        """
        Parameters
        ----------
        coords: tuple
            The coordinates of the point
            The tuple must contain two float coordinates such that the second coordinate is at least as great as the first.
        """
        if len(coords)!=2:
            raise TypeError("Invalid points input. Points in the persistence plane are 2D")
        if coords[0] > coords[1]:
            print(coords)
            raise ValueError("Birth value cannot be greater than death value")
        self._p = tuple(float(x) for x in coords)

    @staticmethod
    def fromstring(s):
        """
        Return a new point object from a string representation.
        """
        return PDPoint([float(x) for x in s.split()])

    def dist(self, other) -> float:
        # The reason this exists as a copy of `self.pp_dist` is to allow easy swapping over to `self.l_inf_dist` if needed.
        """
        Compute quotient distance in the persistence plane.
        The diagonal is treated as a single point.
        """
        return self.pp_dist(other)

    def diagproj(self):
        """
        Compute the projection of this point on the diagonal.
        """
        return PDPoint([(self[0]+self[1])/2, (self[0]+self[1])/2])

    def pp_dist(self, other)->float:
        """
        Compute quotient distance in the persistence plane.
        The diagonal is treated as a single point.
        """
        return min(self.l_inf_dist(other), self.l_inf_dist(self.diagproj())+other.l_inf_dist(other.diagproj()))
    
    def l_inf_dist(self, other)->float:
        """
        Compute the l_inf distance between the `self` and `other`.
        If the dimensions don't match the distance is computed in projection
        down to the common subspace.
        """
        return max(abs(a-b) for (a,b) in zip(self, other))
        
    def isdiagonalpoint(self)->bool:
        """
        Check if this point lies on the diagonal.
        """
        return self[0]==self[1]

    def __getitem__(self, index):
        return self._p[index]

    def __eq__(self, other):
        return self._p == other._p

    def __hash__(self):
        return hash(self._p)

    def __str__(self):
        return " ".join(str(c) for c in self._p)

    def __iter__(self):
        return iter(self._p)

    def __len__(self):
        return len(self._p)

    def __repr__(self):
        return str(self)


from pdsketch import Diagram, DiagramViz
from ds2viz.canvas import svg_plus_pdf
from ds2viz.element import Line, Text
from pdsketch.sketch_style import sketch_style

def plot_matching(a: Diagram, b: Diagram, matching, bottleneck, filename):

    # Select larger of the two diagrams as base
    plot_size_a = max(p[1] for p in a)
    plot_size_b = max(p[1] for p in b)

    base = a if plot_size_a >= plot_size_b else b
    other = a if base == b else b

    base_viz = DiagramViz(base, style='_diagram_a')
    other_viz = DiagramViz(other, style='_diagram_b', plot_size = base_viz.plot_size)

    with svg_plus_pdf(base_viz.plot_size, base_viz.plot_size, filename, sketch_style) as canvas:
        base_viz.draw_skeleton(canvas)

        # Draw every line segment in the matching
        for p_a in matching:
            for p_b in matching[p_a]:
                match_style = '_bottleneck' if p_a.dist(p_b) == bottleneck else '_matching'
                mass = matching[p_a][p_b]

                x = p_b.diagproj() if p_a.isdiagonalpoint() else p_a
                y = p_a.diagproj() if p_b.isdiagonalpoint() else p_b

                Line((x[0], base_viz.plot_size-x[1]), (y[0], base_viz.plot_size-y[1]), style=match_style, stylesheet=sketch_style).draw(canvas)
                label = Text(str(mass), style='_match_mass', stylesheet=sketch_style)
                label.align('left', ((x[0]+y[0])/2-8, base_viz.plot_size - (x[1]+y[1])/2))
                label.draw(canvas)

        # Draw every point
        base_viz.draw_points(canvas)
        other_viz.draw_points(canvas)



from collections import defaultdict
from metricspaces import MetricSpace
from pdsketch import Diagram, PDPoint
from greedypermutation.clarksongreedy import greedy
from collections import namedtuple

SketchEntry = namedtuple('SketchEntry', ['point',
                                         'mass',
                                         'parent',
                                         'update_plan'])

class SketchSequence:
    """
    A class to generate sketches from a persistence diagram.
    """
    def __init__(self, diagram: Diagram = None, n: int = None):
        """
        Parameters
        ----------
        diagram : Diagram
            The PD to be sketched
        n : int
            The number of sketches to be produced
        """
        # If you want an empty sketch, you can have one.
        if diagram is None:
            self._sketches = []
            return

        if n is not None:
            n = len(diagram)

        points, masses = diagram.get_point_mass_lists()
        total_mass = diagram.total_mass
        diagonal = PDPoint([0,0])
        # Adding the diagonal with a multiplicity of 0.
        points.append(diagonal)
        masses.append(0)
        sketches = greedy(M=MetricSpace(points),
                          seed=diagonal,
                          tree=True,
                          gettransportplan=True,
                          mass=masses)

        # Handle the first point
        point, _, update_plan = next(sketches)
        # self._sketches = [SketchEntry(point, 0, None, {point: total_mass})]
        self._sketches = [SketchEntry(point, 0, None, update_plan)]

        # Handle the rest of the points.
        for point, parent, update_plan in sketches:
            mass = diagram.mass[point]
            self._sketches.append(SketchEntry(point, mass, parent, update_plan))
            if len(self._sketches) == n:
                return
        # i = 0
        # while i <= n:
        #     point, parent_index, transport_plan = next(sketches)
        #     if i == 0 or not point.isdiagonalpoint():
        #         # The above condition ensures that only off-diagonal points are added to the sketch except for the 0th sketch which is just the diagonal.
        #         self._sketches.append({'point':point, 'parent_index': parent_index, 'transport_plan': transport_plan.copy()})
        #         i += 1
        # Remove the diagonal projections from the zeroth sketch
        # self._sketches[0]['transport_plan'][diagonal] -= total_mass

    def sketch_bottleneck(self, i:int)->float:
        """
        Returns the bottleneck distance between the original diagram and its ith sketch
        """
        if i >= len(self._sketches) - 1:
            return 0
        else:
            s = self._sketches[i+1]
            parent = self._sketches[s.parent].point
            return s.point.dist(parent)
        # if i < len(self._sketches)-1:
        #     return self._sketches[i+1]['point'].dist(self._sketches[self._sketches[i+1]['parent_index']]['point'])
        # else:
        #     return None

    def _to_dict(str_transport: str):
        """
        Internal method to convert string to transportation plan.
        Used only in method `load_from_file()`.
        """
        transport = defaultdict(int)
        str_transport = str_transport[str_transport.find('{')+1:str_transport.find('}')]
        if str_transport == '':
            return transport
        dict_entries = str_transport.split(", ")
        for entry in dict_entries:
            point_string, mass = entry.split(": ")
            point = PDPoint.fromstring(point_string)
            transport[point] = int(mass)
        return transport

    def load_from_file(filename:str):
        """
        Loads a sketch sequence from a text file.
        File format:
        b_i d_i; parent_i; transportplan_i
        where (b_i, d_i) is the ith point added to the sketch, parent_i is its parent
        and transportplan_i is the ith transportation plan in defaultdict(int) format.
        """
        S = SketchSequence()
        with open(filename, 'r') as s:
            for sketch in s:
                point, mass, parent, update_plan = sketch.rstrip().split("; ")
                point = PDPoint.fromstring(point)
                mass = int(mass)
                parent = int(parent) if parent != 'None' else None
                update_plan = SketchSequence._to_dict(update_plan)
                S._sketches.append(SketchEntry(point, mass, parent, update_plan))
        return S

    def save_to_file(self, filename:str = "sketch.dgm"):
        """
        Save current sketches to a text file.
        The ith line will be saved as follows:
        b_i d_i; mass_i; parent_i; updateplan_i
        where (b_i, d_i) is the ith point added to the sketch, parent_i is the
        index of its parent and updateplan_i is the ith transportation plan as
        a dictionary.
        """
        with open(filename, 'w') as s:
            s.write(str(self))

    def __getitem__(self, index:int)->defaultdict:
        sketch = Diagram()

        for i in range(index+1):
            for p in self._sketches[i].update_plan:
                sketch.add(p, self._sketches[i].update_plan[p])
        return sketch

    def __iter__(self):
        return iter(self._sketches)

    def __len__(self) -> int:
        return len(self._sketches)

    def __str__(self):
        entries = []
        for point, mass, parent, update_plan in self._sketches:
            entries.append(f"{point}; {mass}; {parent}; {dict(update_plan)}")
        return "\n".join(entries)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pdsketch",
    version="0.1.0",
    author="Donald R. Sheehy and Siddharth Sheth",
    author_email="don.r.sheehy@gmail.com",
    description="A Python package for computing sketches of persistence diagrams.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/donsheehy/pdsketch",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=['greedypermutation',
                      'metricspaces',
                     ],
)



from pdsketch import PDPoint
from ds2viz.element import Circle, Text
from pdsketch.sketch_style import sketch_style

class PDPointViz():
    """
    Class to depict a point in the persistence plane with mass using ds2viz.
    """
    def __init__(self, p:PDPoint, mass:int, plot_size:int , style:str=None):
        """
        Parameters
        ----------
        p : PDPoint
            The point to be plotted
        mass : int
            The multiplicity of p
        plot_size : int
            The dimension of the plot.
            Needed to transform p's orientation.
        style : str
            Drawing style.
            Can be a particular diagram color.
        """
        self.point = [p[0], plot_size-p[1]]
        self.mass = mass
        self.style = style if style else '_circle'

    def draw(self, canvas):
        # !!!! Needs a change in ds2viz to run.
        # element.py line 130: change `super().__init__()` to `super().__init__(style, stylesheet)`
        # Have also changed line 134 to modify size of textbox from 17* to 10*

        # Create mass label for a point and draw it
        label = Text(str(self.mass), style='_text', stylesheet=sketch_style)
        label.align('left', (self.point[0], self.point[1]))
        label.draw(canvas)

        # Draw a point in a sketch as a circle
        point = Circle(1, style=self.style, stylesheet=sketch_style)
        point.align('center', (self.point[0], self.point[1]))
        point.draw(canvas)
